#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef __mips__
/* Test and set */
extern int gUSA_test_and_set (volatile int *mem);
__asm__ ("\n\t"
        ".text\n\t"
        ".align 3\n\t"
        ".globl gUSA_test_and_set\n\t"
        ".ent   gUSA_test_and_set\n\t"
        ".set   noreorder\n\t"
"gUSA_test_and_set:\n\t"
       "li      $3, -1\n\t"
       "la      $8, 1f\n\t"
       "move    $9, $sp\n\t"
       "li      $sp, -8         # critical region start: rewind point #1\n"
   "0:  lw      $2, ($4)        # rewind point #2\n\t"
       "sw      $3, ($4)\n"
   "1:  move    $sp, $9         # critical region end\n\t"
       "j       $31\n\t"
       "nop\n\t"
       ".end    gUSA_test_and_set"
);

extern int (*__mips1_tst_func)(volatile int *);
static int
sony_test_and_set (volatile int *mem)
{
  return (*__mips1_tst_func)((volatile int *)mem); 
}

static int
gUSA_exchange_and_add (volatile int *mem, int val)
{
  unsigned long dummy;
  int result;

  __asm__ (".align 2\n\t"
       "la      $8, 1f\n\t"
       "move    $9, $sp\n\t"
       "li      $sp, -12        # critical region start: rewind point #1\n"
   "0:  lw      %0, (%2)        # rewind point #2\n\t"
       "addu    %1, %0, %3\n\t"
       "sw      %1, (%2)\n"
   "1:  move    $sp, $9         # critical region end"
           : "=&r" (result), "=&r" (dummy)
           : "r" (mem), "r" (val)
           : "memory", "$8", "$9");

  return result;
}

static int
TAS_SPIN_SHARE_exchange_and_add (volatile int *mem, int val)
{
  static int lock = 0;
  int result;

  while (gUSA_test_and_set (&lock) != 0)
    asm volatile ("");

  result = mem[0];
  mem[0] = result + val;

  lock = 0;

  return result;
}

static int
TAS_SPIN_DOUBLESPACE_exchange_and_add (volatile int *mem, int val)
{
  static int lock = 0;
  int result;

  while (gUSA_test_and_set (&mem[1]) != 0)
    asm volatile ("");

  result = mem[0];
  mem[0] = result + val;

  mem[1] = 0;

  return result;
}

#define HAVE_SONY_IMPLEMENTATION 1
static int
TAS_SPIN_SHARE_exchange_and_add_sony (volatile int *mem, int val)
{
  static int lock = 0;
  int result;

  while (sony_test_and_set (&lock) != 0)
    asm volatile ("");

  result = mem[0];
  mem[0] = result + val;

  lock = 0;

  return result;
}

static int
TAS_SPIN_DOUBLESPACE_exchange_and_add_sony (volatile int *mem, int val)
{
  static int lock = 0;
  int result;

  while (sony_test_and_set (&mem[1]) != 0)
    asm volatile ("");

  result = mem[0];
  mem[0] = result + val;

  mem[1] = 0;

  return result;
}

#elif defined(__sh__)

static int
gUSA_exchange_and_add (volatile int *mem, int val)
{
  unsigned long dummy;
  int result;

  __asm__ (".align 2\n\t"
       "mova    1f,r0\n\t"
       "nop\n\t"
       "mov     r15,r1\n\t"
       "mov     #-8,r15         ! critical region start: rewind point #1\n"
   "0:  mov.l   @%2,%0          ! rewind point #2\n\t"
       "mov     %3,%1\n\t"
       "add     %0,%1\n\t"
       "mov.l   %1,@%2\n"
   "1:  mov     r1,r15          ! critical region end"
           : "=&r" (result), "=&r" (dummy)
           : "r" (mem), "r" (val)
           : "memory", "r0", "r1");

  return result;
}

static int
TAS_SPIN_SHARE_exchange_and_add (volatile int *mem, int val)
{
  static int lock = 0;
  unsigned long dummy;
  int result;

  __asm__ (".align 2\n\t"
   "0:  tas.b   @%4\n\t"
       "bf      0b\n\t"
       "mov.l   @%3,%0\n\t"
       "add     %0,%1\n\t"
       "mov.l   %1,@%3\n\t"
       "mov     #0,%1\n\t"
       "mov.l   %1,@%4"
           : "=&r" (result), "=r" (dummy)
           : "1" (val), "r" (mem), "r" (&lock)
           : "memory");

  return result;
}

static int
TAS_SPIN_DOUBLESPACE_exchange_and_add (volatile int *mem, int val)
{
  unsigned long dummy;
  int result;

  __asm__ (".align 2\n\t"
   "0:  tas.b   @%4\n\t"
       "bf      0b\n\t"
       "mov.l   @%3,%0\n\t"
       "add     %0,%1\n\t"
       "mov.l   %1,@%3\n\t"
       "mov     #0,%1\n\t"
       "mov.l   %1,@%4"
           : "=&r" (result), "=r" (dummy)
           : "1" (val), "r" (mem), "r" (mem+1)
           : "memory");

  return result;
}

#define HAVE_SYSCALL_eaa 1
static int
SYSCALL_exchange_and_add (volatile int *mem, int val)
{
  register unsigned long __sc3 __asm__ ("r3") = 17;
  register long __sc4 __asm__ ("r4") = (long) mem;
  register long __sc5 __asm__ ("r5") = (long) val;
  unsigned long dummy;
  int result;

  __asm__ __volatile__ ("trapa  #0x12"
           : "=z" (result)
           : "0" (__sc3), "r" (__sc4), "r" (__sc5)
           : "memory");

  if ((unsigned long)(result) >= (unsigned long)(-125)) {
    /* Avoid using "res" which is declared to be in register r0;
       errno might expand to a function call and clobber it.  */
    int __err = -(result);
    errno = __err;
    result = -1;
  }

  return result;
}
#elif defined(__arm__)

static int
gUSA_exchange_and_add (volatile int *mem, int val)
{
  unsigned long dummy;
  int result;

  __asm__ (
       "add	r0, pc, #16\n\t"
       "mov	r1, sp\n\t"
       "mov	sp, #-12	@ critical region start: rewind point #1\n"
   "0:	ldr	%0, [%2]	@ rewind point #2\n\t"
       "add	%1, %0, %3\n\t"
       "str	%1, [%2]\n"
   "1:	mov     sp, r1		@ critical region end"
           : "=&r" (result), "=&r" (dummy)
           : "r" (mem), "r" (val)
           : "memory", "r0", "r1");

  return result;
}

static int
TAS_SPIN_SHARE_exchange_and_add (volatile int *mem, int val)
{
  static int lock = 0;
  unsigned long dummy;
  int result;

  __asm__ (
   "0:  swp	%0, %0, [%4]\n\t"
       "cmp	%0, #0\n\t"
       "bne	0b\n\t"
       "ldr	%0, [%3]\n\t"
       "add	%1, %0, %1\n\t"
       "str	%1, [%3]\n\t"
       "mov	%1, #0\n\t"
       "str	%1, [%4]"
           : "=r" (result), "=r" (dummy)
	   : "1" (val), "r" (mem), "r" (&lock), "0" (1)
           : "memory");

  return result;
}

static int
TAS_SPIN_DOUBLESPACE_exchange_and_add (volatile int *mem, int val)
{
  unsigned long dummy;
  int result;

  __asm__ (".align 2\n\t"
   "0:  swp	%0, %0, [%4]\n\t"
       "cmp	%0, #0\n\t"
       "bne	0b\n\t"
       "ldr	%0, [%3]\n\t"
       "add	%1, %0, %1\n\t"
       "str	%1, [%3]\n\t"
       "mov	%1, #0\n\t"
       "str	%1, [%4]"
           : "=r" (result), "=r" (dummy)
           : "1" (val), "r" (mem), "r" (mem+1), "0" (1)
           : "memory");

  return result;
}
#endif

static int (*eaa)(volatile int *mem, int val);

static int
exchange_and_add (volatile int *mem, int val)
{
  return (*eaa) (mem, val);
}

static int threads_finished = 0;

static int how_many_times;
static int increment;

static int *stack_area;
static int *obj;

static int
countup (void *arg)
{
  volatile int *mem = (volatile int *)arg;
  int i;
  int r;

  for (i = 0; i < how_many_times; i++)
    r = exchange_and_add (mem, increment);

  _exit (0);

  /**/
  return r;
}

static int type;
static int max_object, max_thread;
static int *first_obj;

static void
allocate_objects (void)
{
  int i;

  first_obj = obj = (int *) malloc (max_object * 2 * sizeof (int));
  if (obj == NULL)
    {
      perror ("memory");
      exit (1);
    }
  for (i = 0; i < max_object * 2; i++)
    obj[i] = 0;

  stack_area = (int *) malloc (max_object * max_thread * 4096 + 4096);
  stack_area = (int *) ((((unsigned int) stack_area) + 4095) & ~4095);
  if (stack_area == NULL)
    {
      perror ("memory");
      exit (1);
    }
}

static void
spawn_thread_on_object (int i, int j)
{
  int r;

  r = clone (countup,
             (void *)((unsigned int)stack_area
                      + 4096 * (i * max_thread + j + 1)),
             SIGCHLD|CLONE_FS|CLONE_FILES|CLONE_PTRACE|CLONE_VM,
             &obj[i * 2]);
  if (r < 0)
    {
      perror ("clone");
      exit (3);
    }
}

static void
check_objects (void)
{
  int i, r;
  int bad = 0;

  r = how_many_times * increment * max_thread;

  for (i = 0; i < max_object; i++)
    if (obj[i*2] != r)
      {
        fprintf (stderr, "%d != %d\n", obj[i*2], r);
        bad = 1;
      }

  if (bad)
    exit (2);
}

static void
free_objects (void)
{
  /* Nothing to be done */;
}

static void
sigchld_handler (int signum)
{
  int pid;
  int status;

  if (signal (SIGCHLD, sigchld_handler) == SIG_ERR)
    {
      perror("signal");
      exit(5);
    }

  do
    {
      if ((pid = waitpid (-1, &status, WNOHANG)) < 0)
        {
          if (errno != ECHILD)
            perror("waitpid");
        }

      if (pid > 0)
        {
          threads_finished++;

          if (!WIFEXITED (status))
            {
              fprintf (stderr, "not sane exit\n");
              if (WIFSIGNALED (status))
                fprintf (stderr, "SIG: %d\n", WTERMSIG (status));
            }
          else
            if (WEXITSTATUS (status) != 0)
              fprintf (stderr, "EXIT: %d\n", WEXITSTATUS (status));
        }
    }
  while (pid > 0);
}

#define MAX_BUF 256
static void
sighup_handler (int signum)
{
  char buf[MAX_BUF];
  int n;

  if (signal (SIGHUP, sighup_handler) == SIG_ERR)
    {
      perror("signal");
      exit(5);
    }

  n = snprintf (buf, MAX_BUF, "%d: %d\n", type, *first_obj);

  /* Should be atomic */
  write (1, buf, n);
}

int
main (int argc, char *argv[])
{
  int i, j;
  sigset_t sig_set;
  sigset_t sig_set_old;

  how_many_times = atoi (argv[1]);
  increment = atoi (argv[2]);
  max_object = atoi (argv[3]);
  max_thread = atoi (argv[4]);
  type = atoi (argv[5]);

  if (signal (SIGCHLD, sigchld_handler) == SIG_ERR)
    {
      perror("signal");
      exit(5);
    }

  if (signal (SIGHUP, sighup_handler) == SIG_ERR)
    {
      perror("signal");
      exit(5);
    }

  switch (type)
    {
    case 0:
      eaa = gUSA_exchange_and_add;
      break;
    case 1:
      eaa = TAS_SPIN_SHARE_exchange_and_add;
      break;
    case 2:
      eaa = TAS_SPIN_DOUBLESPACE_exchange_and_add;
      break;
#ifdef HAVE_SYSCALL_eaa
    case 3:
      eaa = SYSCALL_exchange_and_add;
      break;
#endif
#ifdef HAVE_SONY_IMPLEMENTATION
    case 4:
      eaa = TAS_SPIN_SHARE_exchange_and_add_sony;
      break;
    case 5:
      eaa = TAS_SPIN_DOUBLESPACE_exchange_and_add_sony;
      break;
#endif
    default:
      _exit (1);
    }

  allocate_objects ();

  for (i = 0; i < max_object; i++)
    for (j = 0; j < max_thread; j++)
      spawn_thread_on_object (i, j);

  sigemptyset (&sig_set);
  sigaddset (&sig_set, SIGCHLD);
  sigprocmask (SIG_BLOCK, &sig_set, &sig_set_old);
  while (threads_finished < (max_thread * max_object))
    sigsuspend (&sig_set_old);

  check_objects ();

  free_objects ();

  return 0;
}
