#!/usr/bin/env python

import sys, string, popen2, signal, os
from qt import *

TRUE  = 1
FALSE = 0

class MyFileHandler(QSocketNotifier):
    def __init__(self, file, progressbar):
        QSocketNotifier.__init__(self, file.fileno(), QSocketNotifier.Read, progressbar)
        self.progress = 0
        self.file = file
        self.progressbar = progressbar

class MyWidget(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self,None,'gUSA Demonstration',Qt.WDestructiveClose)

        self.filename = QString.null

        self.file = QPopupMenu(self)
        self.menuBar().insertItem('&File',self.file)
        self.file.insertItem('&Quit',qApp,SLOT('closeAllWindows()'),Qt.CTRL + Qt.Key_Q)

        self.help = QPopupMenu(self)
        self.menuBar().insertSeparator()
        self.menuBar().insertItem('&Help',self.help)
        self.help.insertItem('&About',self.about,Qt.Key_F1)

        self.qv = QVBox(self)
        self.setCentralWidget(self.qv)

        self.qh = QHBox(self.qv)
        self.qh.setSpacing(5)

        self.gb0 = QGroupBox(2, Qt.Vertical, self.qh)
        QLabel("Count    ",self.gb0)
        self.count = QSpinBox(0, 1000000, 100, self.gb0)
        self.count.setValue(300000)
        QLabel("Objects",self.gb0)
        self.objects = QSpinBox(self.gb0)
        self.objects.setMinValue(1)
        self.objects.setMaxValue(64)
        QLabel("Threads(per obj.)",self.gb0)
        self.threads = QSpinBox(self.gb0)
        self.threads.setMinValue(1)
        self.threads.setMaxValue(64)

        self.qv0 = QVBox(self.qh)
        self.qv0.setSpacing(10)
        self.startbutton = QPushButton("Start",self.qv0)
        self.connect(self.startbutton, SIGNAL('clicked()'), self.start)
        self.stopbutton = QPushButton("Stop",self.qv0)
        self.connect(self.stopbutton, SIGNAL('clicked()'), self.stop)

        self.gb1 = QGroupBox(3, Qt.Horizontal, self.qv)

        QLabel("gUSA",self.gb1)
        self.progressbar0 = QProgressBar(self.gb1)
        self.result0 = QLabel("      ", self.gb1)
        QLabel("Shared",self.gb1)
        self.progressbar1 = QProgressBar(self.gb1)
        self.result1 = QLabel("      ", self.gb1)
        QLabel("Distrib.",self.gb1)
        self.progressbar2 = QProgressBar(self.gb1)
        self.result2 = QLabel("      ", self.gb1)

        self.status = QStatusBar(self.qv)

    def about(self):
        QMessageBox.about(self,'gUSA Demonstration',
            'This demonstrates gUSA,\nSimple and Efficient\nUser Space Atomicity Emulation.')

    def interval(self):
        self.time = self.time + 0.1
        if (self.childstop == 3):
            self.t.stop()
        else:
            if (self.process0 and self.process0.poll() == -1):
                os.kill(self.process0.pid, signal.SIGHUP)
            if (self.process1 and self.process1.poll() == -1):
                os.kill(self.process1.pid, signal.SIGHUP)
            if (self.process2 and self.process2.poll() == -1):
                os.kill(self.process2.pid, signal.SIGHUP)

    def start(self):
        self.childstop = 0
        self.progressbar0.setTotalSteps(self.count.value())
        self.progressbar1.setTotalSteps(self.count.value())
        self.progressbar2.setTotalSteps(self.count.value())
        self.progressbar0.setProgress(0)
        self.progressbar1.setProgress(0)
        self.progressbar2.setProgress(0)
        self.result0.setText("      ")
        self.result1.setText("      ")
        self.result2.setText("      ")
        self.status.message("Start!")
        self.t = QTimer(self, "timer")
        self.connect(self.t, SIGNAL('timeout()'), self.interval)
        self.t.start(100) # 100ms
        self.time = 0.0

        t = QTimer(self, "oneshot")
        t.start(800, TRUE)
        self.connect(t, SIGNAL('timeout()'), self.message_clear)

        self.process0 = popen2.Popen3("./eaa " + `self.count.value()`  + " 1 " + `self.objects.value()` + " " + `self.threads.value()` + " 10000");
        self.process1 = popen2.Popen3("./eaa " + `self.count.value()`  + " 1 " + `self.objects.value()` + " " + `self.threads.value()` + " 15000");
        self.process2 = popen2.Popen3("./eaa " + `self.count.value()`  + " 1 " + `self.objects.value()` + " " + `self.threads.value()` + " 30000");

        # PyQt's QSocketNotifier is buggy on signal setting???
        self.fh0 = MyFileHandler(self.process0.fromchild, self.progressbar0)
        self.connect(self.fh0, SIGNAL('activated(int)'), self.dataready_fh0)
        self.fh1 = MyFileHandler(self.process1.fromchild, self.progressbar1)
        self.connect(self.fh1, SIGNAL('activated(int)'), self.dataready_fh1)
        self.fh2 = MyFileHandler(self.process2.fromchild, self.progressbar2)
        self.connect(self.fh2, SIGNAL('activated(int)'), self.dataready_fh2)

    def dataready_fh0(self,fileno):
        if (self.fh0):
            str = self.fh0.file.readline()
            if (str):
                self.fh0.progress = int(str)
                self.fh0.progressbar.setProgress(self.fh0.progress)
            else:
                self.fh0.setEnabled(FALSE)
                if (self.process0 and self.process0.poll() == -1):
                    os.kill(self.process0.pid, signal.SIGKILL)
                self.fh0.progressbar.setProgress(self.fh0.progressbar.totalSteps())
                self.status.message("gUSA done!")
                t = QTimer(self, "oneshot")
                t.start(2000, TRUE)
                self.connect(t, SIGNAL('timeout()'), self.message_clear)
                self.result0.setNum(self.time)
                self.fh0 = 0
                self.process0 = 0
                self.childstop = self.childstop + 1

    def dataready_fh1(self,fileno):
        if (self.fh1):
            str = self.fh1.file.readline()
            if (str):
                self.fh1.progress = int(str)
                self.fh1.progressbar.setProgress(self.fh1.progress)
            else:
                self.fh1.setEnabled(FALSE)
                if (self.process1 and self.process1.poll() == -1):
                    os.kill(self.process1.pid, signal.SIGKILL)
                self.fh1.progressbar.setProgress(self.fh1.progressbar.totalSteps())
                self.status.message("Shared done!")
                t = QTimer(self, "oneshot")
                t.start(2000, TRUE)
                self.connect(t, SIGNAL('timeout()'), self.message_clear)
                self.result1.setNum(self.time)
                self.fh1 = 0
                self.process1 = 0
                self.childstop = self.childstop + 1

    def dataready_fh2(self,fileno):
        if (self.fh2):
            str = self.fh2.file.readline()
            if (str):
                self.fh2.progress = int(str)
                self.fh2.progressbar.setProgress(self.fh2.progress)
            else:
                self.fh2.setEnabled(FALSE)
                if (self.process2 and self.process2.poll() == -1):
                    os.kill(self.process2.pid, signal.SIGKILL)
                self.fh2.progressbar.setProgress(self.fh2.progressbar.totalSteps())
                self.status.message("Distrib. done!")
                t = QTimer(self, "oneshot")
                t.start(2000, TRUE)
                self.connect(t, SIGNAL('timeout()'), self.message_clear)
                self.result2.setNum(self.time)
                self.fh2 = 0
                self.process2 = 0
                self.childstop = self.childstop + 1

    def message_clear(self):
        self.status.message("")

    def stop(self):
        self.status.message("Stopped")
        self.t.stop()
        t = QTimer(self, "oneshot")
        t.start(800, TRUE)
        self.connect(t, SIGNAL('timeout()'), self.message_clear)
        if (self.process0.poll() == -1):
            self.fh0.setEnabled(FALSE)
            os.kill(self.process0.pid, signal.SIGKILL)
            self.fh0 = 0
            self.process0 = 0
        if (self.process1.poll() == -1):
            self.fh1.setEnabled(FALSE)
            os.kill(self.process1.pid, signal.SIGKILL)
            self.fh1 = 0
            self.process1 = 0
        if (self.process2.poll() == -1):
            self.fh2.setEnabled(FALSE)
            os.kill(self.process2.pid, signal.SIGKILL)
            self.fh2 = 0
            self.process0 = 0

a = QApplication(sys.argv)

w = MyWidget()
a.setMainWidget(w)
w.show()
a.exec_loop()
