#!/bin/sh
#
# script for generating dxs filesystem from existing linux installation
#
VERSION="0.5.1e"
#
# We assume four nfs shares exported:
# 1) the root filesystem without /usr, /opt, /tmp/users which is created
#    here
# 2) /usr, /opt directly exported from server filesystem
# 3) /tmp/users as a special rw exported share for temporary files
#
# Dirk von Suchodoletz <dirk@goe.net>, 11-02-2003
#
###########################################################################

# helper functions
#
# compute the subnet address of the local network
ipv4_subnet() {
  local ip="$1"
  local netmask="$2"

  # Split quad-dotted addresses into bytes
  # There is no double quote around the back-quoted expression on purpose
  # There is no double quote around $ip and $netmask on purpose
  set `IFS='.'; echo $ip $netmask`

  echo $(($1 & $5)).$(($2 & $6)).$(($3 & $7)).$(($4 & $8))
}
# compute the broadcast address
ipv4_broadcast() {
  local ip="$1"
  local netmask="$2"

  # Split quad-dotted addresses into bytes
  # There is no double quote around the back-quoted expression on purpose
  # There is no double quote around $ip and $netmask on purpose
  set `IFS='.'; echo $ip $netmask`

  echo $(($1 | (255 - $5))).$(($2 | (255 - $6))).$(($3 | (255 - $7))).$(($4 | (2
55 - $8)))
}

#
# declare some variables and set some defaults
NFSROOT="/nfsroot"
NETNAME=`route -n | grep eth0 | grep -v "UG" | awk '{ print $1 }'`
NETMASK=`route -n | grep eth0 | grep -v "UG" | awk '{ print $3 }'`
#NETMASK=`ifconfig eth0|grep "inet addr"|awk -F : '{print $4 }'|awk '{print $1}'`
SERVER=`ifconfig eth0|grep "inet addr"|awk -F : '{print $2 }'|awk '{print $1}'`
if [ -z $NETNAME ] ; then
	NET="192.168.2.0/255.255.255.0"
else
	NET="$NETNAME/$NETMASK"
fi
PROGRESS=0
ETCSUBDIRS="/etc/cron.daily /etc/cups /etc/default /etc/java /etc/profile.d \
/etc/unixODBC /etc/gimp /etc/gtk /etc/pam.d /etc/skel /etc/iproute2 /etc/xml \
/etc/WindowMaker /etc/ssl /etc/security /etc/alsa.d /etc/gpm /etc/openldap \
/etc/opt /etc/postfix /etc/ssh /etc/cron.d /etc/sane.d /etc/libiodbc \
/etc/radiusclient /etc/diameter /etc/raddb /etc/ipsec.d /etc/texmf \
/etc/samba /etc/ximian /etc/stunnel /etc/hotplug /etc/vmware \
/etc/SuSEconfig"
ETCFILES="bash.bashrc profile protocols shells ttytype vimrc wgetrc \
enscript.cfg esd.conf zshenv zshrc networks netgroup pluggerrc mesa.conf \
hosts.allow hosts.deny hosts.equiv securetty codecs.conf input.conf \
mplayer.conf fam.conf lesskey lesskey.bin mime.types raw rpc screenrc"

# check files and directories for existence
TMPVAR=""
for check in $ETCSUBDIRS; do
	test -d $check && TMPVAR=$TMPVAR" "$check;
done
ETCSUBDIRS=$TMPVAR
TMPVAR=""
for check in $ETCFILES; do
	test -f /etc/$check && TMPVAR=$TMPVAR" "$check;
done
ETCFILES=$TMPVAR

(( NARG=0 ))
TEST=( $@ )
while [ $NARG -le $# ] ; do
	case "${TEST[$NARG]}" in
	*help*)
		echo -n "$0 generates dxs filesystem from existing "
		echo -ne "linux installation\nPass nfsroot with the"
		echo -e "$0 -nfsroot <value> option\n"
		exit 0
	;;
	*nfsroot*)
		NFSROOT=${TEST[$NARG+1]}
		(( NARG=$NARG+1 ))
	;;
	*exp*)
		NET=${TEST[$NARG+1]}
		(( NARG=$NARG+1 ))
	;;
	esac
(( NARG=$NARG+1 ))
done

echo "Welcome to version $VERSION of mk_dxsfs.sh!"
echo
echo "Please answer the following questions: (Enter takes defaults)"
echo
echo "Which network do you want to use for DXS? (A.B.C.0)"
echo -n "* [ $NETNAME ] "
read NN
if [ -z NN ] ; then NETNAME=NN; fi
echo "Which netmask should be used? (255.B.C.0)"
echo -n "* [ $NETMASK ] "
read NM
if [ -z NM ] ; then NETMASK=NN; fi
echo "Using $NETNAME/$NETMASK !"
KERNEL=3
echo "Which kernel should be used?"
echo "Dont forget to match your dhcpd.conf settings to the appropriate"
echo "framebuffer resolutions: lpp->0x301, splash->0x317!"
echo "  [1] 2.4.18 splash"
echo "  [2] 2.4.18 nosplash"
echo "* [3] 2.4.19 splash"
echo "  [4] 2.4.19 nosplash"
echo "  [5] 2.4.19 lpp"
read KRN
if [ -z KRN ] ; then KERNEL=KRN; fi
echo "Where automount home directories from? (A.B.C.D:/home-dir)"
echo -n "* [ $SERVER:/home ]"
read AM
if [ -z AM ] ; then AMT=AM; 
else AMT=$SERVER:/home; fi

echo "This script sets up the basic filesystem structure for diskless"
echo "X-stations now. Please enable the NFS-Share $NFSROOT/dxs in your"
echo "servers /etc/exports file, i.e.:"
echo "$NFSROOT/dxs 	$NET(ro,no_root_squash)"
echo "/tmp/dxs	$NET(rw,no_root_squash)"
echo "/usr		$NET(ro)"
echo "/opt		$NET(ro)"
echo
echo "If you would like to use TeX too enable the following lines"
echo "/var/lib/texmf $NET(ro)"
echo "/var/cache/fonts $NET(rw)"
echo
echo "... and (re)start your nfs server."
echo

# create mountpoint for dxs tmp directory
echo -e "Creating dxs tmp directory in '/tmp/dxs'. Do not delete it! Change\
\nmountpoint in dhclient-script instead if you like to place it elsewhere!\n"
mkdir -m 1777 /tmp/dxs || { echo -e "Failed to create /tmp/dxs directory \
for diskless X stations!\n(remove old entries before.)"; exit 1; }

# checking for directory entries to create
test -f exclude || { echo "File with exclude list for DXS filesystem\
 not found!"; exit 1; }
# build an exclude list - extract the directories from 'exclude'
DIR=''; for i in `cat exclude` ; do if [ -d $i ] ; then if [ $DIR ] ;
then DIR="$DIR|(^$i)"; else DIR="(^$i)"; fi ; fi ; done
# exclude backup files too
if [ -z $DIR  ] ; then
  DIR="*.*save|*.*bak|*.*old";
  else DIR="$DIR|*.*save|*.*bak|*.*old";
fi
# exclude the files (hardlinks) from 'exclude'
HL=''; for i in `cat exclude` ; do if [ -f $i ] ; then if [ $HL ] ;
then HL="$HL|($i)"; else HL="($i)"; fi ; fi ;  done
# exclude backup files too
if [ -z $HL ] ; then
  HL="*.*save|*.*bak|*.*old|*~";
  else HL="$HL|*.*save|*.*bak|*.*old|*~";
fi
# exclude the softlinks listed in 'exclude'
SL=''; for i in `cat exclude` ; do if [ -L $i ] ; then if [ $SL ] ;
then SL="$SL|($i)"; else SL="($i)"; fi ; fi ; done
if [ -z $SL ] ; then
  SL="empty-string";
fi

# untarring basic filesystem structure
echo -n "creating basic filesystem structure: "
test -f dxs.tgz || { echo "File with basic filesystem structure for DXS \
 not found!"; exit 1; }
test -d $NFSROOT || mkdir -p $NFSROOT || 
	{ echo "problems creating nfsroot -> $NFSROOT"; exit 1; }
echo -n ".."
for i in `tar -xvpzf dxs.tgz -C $NFSROOT` ; do
	(( PROGRESS=$PROGRESS+1 ))
	if [ $PROGRESS -ge 10 ] ; then echo -n "." ; PROGRESS=0; fi
done
echo " done"

# fake link for easier configuration (removed later)
ln -fs $NFSROOT/dxs/etc.s $NFSROOT/dxs/etc

# creating directories
echo -n "creating directories: " 
for i in `find /bin /sbin /lib $ETCSUBDIRS -type d 2>/dev/null | \
	grep -E -v $DIR` ; do
	(( PROGRESS=$PROGRESS+1 ))
	if [ $PROGRESS -ge 10 ] ; then echo -n "." ; PROGRESS=0; fi 
	test -d $NFSROOT/dxs$i || test -L $NFSROOT/dxs$i \
	|| mkdir -p $NFSROOT/dxs$i
done
echo " done"

# hardlinking or copying files
CMD="ln"
DO="hardlinking"
ln /bin/mount $NFSROOT/dxs/bin/mount &>/dev/null || \
	{ CMD="cp -a"; DO="copying"; } 
echo -n "$DO files: " 
for i in `find /bin /sbin /lib $ETCSUBDIRS -type f | grep -v -E "$DIR" \
	| grep -E -v $HL` ; do
	(( PROGRESS=$PROGRESS+1 ))
	if [ $PROGRESS -ge 10 ] ; then echo -n "." ; PROGRESS=0; fi 
	test -f $NFSROOT/dxs$i || test -L $NFSROOT/dxs$i \
	|| test -e $NFSROOT/dxs$i || $CMD $i $NFSROOT/dxs$i
done
echo " done"

# softlinking files
echo -n "adding soft links: "
for i in `find /bin /sbin /lib $ETCSUBDIRS -type l | grep -E -v $DIR\
	| grep -E -v $SL` ; do
	(( PROGRESS=$PROGRESS+1 ))
	if [ $PROGRESS -ge 10 ] ; then echo -n "." ; PROGRESS=0; fi 
	test -L $NFSROOT/dxs$i || cp -a $i $NFSROOT/dxs$i
done
echo " done"

# softlinking /etc subdirs
for i in $ETCSUBDIRS ; do
	i=`echo $i | sed -e s,/etc/,,i`
	test -L $NFSROOT/dxs/var/ram/etc/$i || \
		ln -s /etc.s/$i $NFSROOT/dxs/var/ram/etc/$i
done

# copy the chooser binary
cp /etc/X11/xdm/chooser $NFSROOT/dxs/etc.s/X11/xdm

# hardlinking some standard /etc files (add all the other to the ETCFILES
# variable you will need for your applications)
for i in $ETCFILES ; do
	ln -f /etc/$i $NFSROOT/dxs/etc.s/$i
	test -L $NFSROOT/dxs/var/ram/etc/$i || \
	ln -s /etc.s/$i $NFSROOT/dxs/var/ram/etc/$i 2>/dev/null
done

# touch the modules.dep files
touch $NFSROOT/dxs/lib/modules/*/modules.* &

# change fake link back to original one
rm $NFSROOT/dxs/etc
ln -fs /RAM/etc $NFSROOT/dxs/etc

# configuring automounter for home directories
echo -e "# /etc/auto.home\n#\n# created by mk_dxsfs.sh $VERSION\n" \
    >$NFSROOT/dxs/etc.s/auto.home
echo -e "*\t-rsize=8192,wsize=8192,rw\t$AMT/&" \
    >>$NFSROOT/dxs/etc.s/auto.home

echo "Please check all configuration files for correctness!"

#echo "Merging boot kernel with the ramdisk file"
#[ -f $dxsdir/boot/bootimg ] && \
#	cp $dxsdir/boot/bootimg $dxsdir/boot/bootimg.old
#MKNBI=`which mknbi-linux`
#[ x$MKNBI="x" ] && { echo "Command 'mknbi-linux' not installed."; exit 1; }
#mknbi-linux --output=$NFSROOT/dxs/boot/bootimg \
#	$NFSROOT/dxs/boot/bzImage inittar.dxs

# do some sanity checks
echo "Doing some checks on your installation ..."
grep -s "$NFSROOT/dxs.*no_root_squash" /etc/exports &>/dev/null || \
echo -e "->> Problem occured: <<-\n\
You should have an entry for the diskless client root filesystem \
\nlooking like:\n $NFSROOT/dxs 1.2.3.4/255.255.255.0(ro,no_root_squash)\n\
Do not forget the 'no_root_squash' option! It is important for running \
\ncommands (like devfsd) with root priviliges!"
grep -s "/usr" /etc/exports &>/dev/null || \
echo -e "->> Problem occured: <<-\n\
You should permit your clients to use your servers '/usr'!"
grep -s "/opt" /etc/exports &>/dev/null || \
echo -e "->> Problem occured: <<-\n\
You should permit your clients to use your servers '/opt'!"
[ -x $NFSROOT/dxs/sbin/devfsd ] || \
echo -e "->> Problem occured: <<-\n\
Could not find the 'devfsd' (device filesystem daemon) in \
$NFSROOT/dxs/sbin!\n You will need it to properly run your machines."
[ -x $NFSROOT/dxs/sbin/dhclient ] || \
echo -e "->> Problem occured: <<-\n\
Could not find the 'dhclient' programm to get configuration data \
from\n dhcp servers! You do not will succeed booting your diskless clients."
which mknbi-linux &>/dev/null || \
echo -e "->> Problem occured: <<-\n\
Could not find the 'mknbi-linux' programm! It is needed to tag the boot \
kernel for use with etherboot."
which gethostip &>/dev/null || \
echo -e "->> Problem occured: <<-\n\
Could not find the 'gethostip' programm! It may be needed for pxelinux.cfg \
configuration file naming. May be ignored if etherboot is the only choice."

# generate pxelinux.cfg network configuration file (quick hack!)
PXECFG="boot/pxelinux.cfg/"`gethostip $NETNAME|awk '{print$3}'||echo default`
echo -e "# pxelinux.cfg/$NETNAME file\n# for instructions see syslinux \
package by Peter Anvin\n#\n# Setup for linux diskless clients in network \
$NETNAME\n#\n# Dirk von Suchodoletz <dirk@goe.net>, 20-01-2003\n#\n \
# created by mk_dxsfs.sh $VERSION\n" > $NFSROOT/dxs/$PXECFG
echo -e "label linux\n\tkernel bzImage" >>$NFSROOT/dxs/$PXECFG

# link bzImage to the kernel file choosen
RPT=`pwd`
cd $NFSROOT/dxs/boot
case $KERNEL in
    1|2)
	ln -sf bzImage-2.4.18 bzImage
	echo -e "\tappend initrd=inittar.nosplash">>$NFSROOT/dxs/$PXECFG
	;;
    3|4)
	ln -sf bzImage-2.4.19 bzImage
	echo -e "\tappend vga=0x317 initrd=inittar.splash debug=0">>$NFSROOT/dxs/$PXECFG
	;;
    5)
  	ln -sf bzImage-2.4.19-lpp bzImage
	echo -e "\tappend vga=0x301 initrd=inittar.nosplash">>$NFSROOT/dxs/$PXECFG
	;;
esac
cd $RPT

# setting up boot kernel dependend on SuSE version
if test -f /etc/SuSE-release -a -f /bin/splash ; then
   	VER=`cat /etc/SuSE-release | grep VERSION | sed -e "s,.*= ,,g"`
	case "$VER" in
       	8.0|8.1)
	    echo "Creating INITTAR file now (8.X versions) ..."
   	    echo "Running 'mkdxsinittar -dxsrd \"$NFSROOT/dxs\" -k \"bzImage\"' ..."
	    ./mkdxsinittar -dxsrd "$NFSROOT/dxs" -k "bzImage"
		;;
	*)
	    echo "Creating INITTAR file now (7.3 and older) ..."
	    echo "Running 'mkdxsinittar.old -dxsrd \"$NFSROOT/dxs\" -k \"bzImage\"' ..."
	    ./mkdxsinittar.old -dxsrd "$NFSROOT/dxs" -k "bzImage" 
	;;
	esac
else
   	echo "Creating INITTAR file now without splash"
	echo "Running 'mkdxsinittar -dxsrd \"$NFSROOT/dxs\" -k \"bzImage\"' ..."
	./mkdxsinittar -dxsrd "$NFSROOT/dxs" -k "bzImage"
fi

echo "... done"
exit 0
